@@ -26,16 +26,14 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: draw.c,v 1.151 1999/06/08 02:43:58 nishida Exp $
+ * $Id: draw.c,v 1.169 2000/03/05 07:26:30 nishida Exp $
  */
 
 #include "mgp.h"
 
 /* state associated with the window - how should we treat this? */
-static struct ctrl *bg_image_ctl[2] = {NULL, NULL};
-static struct ctrl *bg_image_prev[2] = {NULL, NULL};
-static Image *bg_image[2] = {NULL, NULL};
-static XImageInfo *bg_ximageinfo[2] = {NULL, NULL};
+static struct ctrl *bg_ctl, *bg_ctl_last, *bg_ctl_cache;
+static int bgindex = 0;
 struct render_state cache_state;
 
 static u_short kinsokutable[] = {
@@ -56,9 +54,11 @@ static struct pcache {
 } pcache;
 
 #define	COMPLEX_BGIMAGE \
-    (bg_image_prev[caching]				\
-     && ((bg_image_prev[caching]->ct_op == CTL_BIMAGE)	\
-       || bg_image_prev[caching]->ct_op == CTL_BGRAD))
+    (bg_ctl		\
+     && ((bg_ctl->ct_op == CTL_BIMAGE)	\
+       || bg_ctl->ct_op == CTL_BGRAD))
+
+#define	COMPLEX_BGIMAGE2 (0)
 
 #define	POSY(size)	(-(int)((size)/2))
 
@@ -118,14 +118,15 @@ static void x_registerseed __P((struct render_state *, char *, char *));
 static char *x_findseed __P((struct render_state *, char *));
 
 static void XClearPixmap __P((Display *, Drawable));
-static void XSetPixmapBackgroundPixmap __P((Display *, Drawable, Pixmap, 
-	int, int));
-static void XSetPixmapBackground __P((Display *, Drawable, u_long));
 static void cache_page __P((struct render_state *, int));
 static void cache_effect1 __P((void));
 static void cache_effect2 __P((void));
 static void set_from_cache __P((struct render_state *));
 static void pcache_process __P((int));
+static void predraw __P((struct render_state *));
+static void set_background_pixmap __P((struct ctrl *));
+static void get_background_pixmap __P((struct ctrl *, struct render_state *));
+static void regist_background_pixmap __P((XImageInfo *, Image *));
 
 #define CHECK_CACHE {if (caching){caching = -1; return;}}
 
@@ -159,18 +160,24 @@ state_goto(state, page, repaint)
 	if (!repaint)
 		purgechild(state->page);
 
-	free_alloc_colors(&image_clr);
-	free_alloc_colors(&font_clr);
 	state->page = page;
 	state->line = 0;
 	state->cp = NULL;
 	state->phase = P_NONE;
+	free_alloc_colors(&image_clr);
+	free_alloc_colors(&font_clr);
+
+#ifdef COLOR_BUGFIX
+	colormap = XCopyColormapAndFree(display, colormap);
+#endif
+	predraw(state);
 }
 
 void
 state_next(state)
 	struct render_state *state;
 {
+
 	switch (state->phase) {
 	case P_NONE:
 		fprintf(stderr, "internal error\n");
@@ -247,6 +254,13 @@ state_newpage(state)
 {
 	state->ypos = 0;
 	state->have_mark = 0;
+	free_alloc_colors(&image_clr);
+	free_alloc_colors(&font_clr);
+
+#ifdef COLOR_BUGFIX
+	colormap = XCopyColormapAndFree(display, colormap);
+#endif
+	predraw(state);
 }
 
 /*
@@ -491,7 +505,8 @@ process_direc(state, seenpause)
 		break;
 
 	case CTL_BACK:
-		bg_image_ctl[caching] = cp;	/*update later*/
+		back_color[caching] = cp->ctl_value;
+		bg_ctl = cp;	/*update later*/
 		break;
 
 	case CTL_CCOLOR:
@@ -565,13 +580,13 @@ process_direc(state, seenpause)
 	case CTL_BIMAGE:
 		if (mgp_flag & FL_BIMAGE)
 			break;
-		bg_image_ctl[caching] = cp;	/*update later*/
+		bg_ctl = cp;	/*update later*/
 		break;
 
 	case CTL_BGRAD:
 		if (mgp_flag & FL_BIMAGE)
 			break;
-		bg_image_ctl[caching] = cp;	/*update later*/
+		bg_ctl = cp;	/*update later*/
 		break;
 
 	case CTL_LCUTIN:
@@ -620,44 +635,38 @@ process_direc(state, seenpause)
 
 	case CTL_LINESTART:
 		if (state->line == 0) {
-			if (bg_image_ctl[caching]) {
-				if (bg_image_prev[caching]
-				 && ctlcmp(bg_image_ctl[caching], 
-					bg_image_prev[caching]) == 0) {
-					; /* same as the last time */
-				} else if (bg_image_ctl[caching]->ct_op == CTL_BIMAGE) {
-				    image_load(state, bg_image_ctl[caching]->ctm_fname,
-					bg_image_ctl[caching]->ctm_numcolor,
-					bg_image_ctl[caching]->ctm_ximagesize,
-					bg_image_ctl[caching]->ctm_yimagesize, 1,
-					bg_image_ctl[caching]->ctm_zoomflag, 0);
-				} else if (bg_image_ctl[caching]->ct_op == CTL_BGRAD) {
-					back_gradation(state,
-						&bg_image_ctl[caching]->ct_val.ctrl_grad);
-				} else if (bg_image_ctl[caching]->ct_op == CTL_BACK) {
-					if (!caching)
-						XSetWindowBackground(display,
-							state->target, bg_image_ctl[caching]->ctl_value);
-					else {
-						XSetPixmapBackground(display,
-							state->target, bg_image_ctl[caching]->ctl_value);
+			/*
+			 * set background of target 
+			 */
+			if (bg_ctl) {
+				if (!caching){
+					/* target is window, so we need care bg_ctl_last */
+					if (bg_ctl_last && !ctlcmp(bg_ctl, bg_ctl_last)){
+						/* same as last time, we do nothing  */
+						;
+					} else {
+						/* we have to change background */
+						get_background_pixmap(bg_ctl, state);
+
+						/* set window background */
+						set_background_pixmap(bg_ctl);
+
+						bg_ctl_last = bg_ctl;
 					}
-					back_color[caching] = bg_image_ctl[caching]->ctl_value;
+					XClearWindow(display, state->target);
 				} else {
-					fprintf(stderr, "internal error: "
-						"bogus background\n");
-					cleanup(-1);
+					get_background_pixmap(bg_ctl, state);
+					bg_ctl_cache = bg_ctl;
+
+					XClearPixmap(display, state->target);
 				}
+			} else {
+				if (!caching)
+					XClearWindow(display, state->target);
+				else
+					XClearPixmap(display, state->target);
 			}
 
-			if (!caching) 
-				XClearWindow(display, state->target);
-			else 
-				XClearPixmap(display, state->target);
-
-			bg_image_prev[caching] = bg_image_ctl[caching];
-			bg_image_ctl[caching] = NULL;
-
 			if (t_fin)
 				timebar(state);
 		}
@@ -713,6 +722,10 @@ process_direc(state, seenpause)
 	case CTL_TFONT0:
 		freetypefont0 = cp->ctc_value;
 		break;
+
+	case CTL_TMFONT0:
+		freetypemfont0 = cp->ctc_value;
+		break;
 #endif
 
 	case CTL_NOOP:
@@ -1205,8 +1218,8 @@ draw_fragment(state, p, len, registry, charset16)
 				fprintf(stderr, "can't load font size %d "
 					"(nor font in similar size) for "
 					"font <%s:%d:%s>, glyph 0x%04x\n",
-					char_size, x_findseed(state, registry),
-					char_size, registry, code);
+					char_size[caching], x_findseed(state, registry),
+					char_size[caching], registry?registry:"NULL", code);
 			}
 			break;
 		}
@@ -1329,7 +1342,7 @@ obj_free(state, obj)
 		freeImage(obj->data.image.image);
 		break;
 	case O_XFONT:
-		/* do nothing */
+		free(obj->data.xfont.xfont);
 		break;
 	case O_ICON:
 		if (obj->data.icon.xpoint)
@@ -1408,7 +1421,7 @@ obj_new_xfont(state, x, y, size, code, registry)
 	obj->y = y;
 	obj->fore = fore_color[caching];
 	obj->type = O_XFONT;
-	obj->data.xfont.xfont = x_findseed(state, registry);
+	obj->data.xfont.xfont = strdup(x_findseed(state, registry));
 	obj->data.xfont.csize = size;
 	obj->data.xfont.code = code;
 	obj->data.xfont.registry = registry;
@@ -1603,8 +1616,8 @@ obj_image_trans(image, x, y)
 		bw = bh = 0;	/* for lint */
 		goto end;
 	}
-	bw = bg_image[caching]->width;
-	bh = bg_image[caching]->height;
+	bh = bgpixmap[bgindex].image->height;
+	bw = bgpixmap[bgindex].image->width;
 	j = 0;
 	if (image->type == IBITMAP) {
   expand:
@@ -1617,7 +1630,7 @@ obj_image_trans(image, x, y)
 	}
 	pl = image->pixlen;
 	p = image->data + image->width * j * pl;
-	bpl = bg_image[caching]->pixlen;
+	bpl = bgpixmap[bgindex].image->pixlen;
 	pd = -1;
 	n = 0;	/* for lint */
 	trans = image->trans;
@@ -1625,19 +1638,21 @@ obj_image_trans(image, x, y)
 	for ( ; j < image->height; j++) {
 		by = (y + j) % bh;
 		bx = x % bw;
-		b = bg_image[caching]->data + (bg_image[caching]->width * by + bx) * bpl;
+		b = bgpixmap[bgindex].image->data + 
+			(bgpixmap[bgindex].image->width * by + bx) * bpl;
 		for (i = 0; i < image->width; i++, p += pl, b += bpl, bx++) {
 			if (bx == bw) {
 				bx = 0;
-				b = bg_image[caching]->data + 
-					bg_image[caching]->width * by * bpl;
+				b = bgpixmap[bgindex].image->data + 
+					bgpixmap[bgindex].image->width * by * bpl;
 			}
 			if (memToVal(p, pl) != trans)
 				continue;
 			d = memToVal(b, bpl);
 			if (d != pd) {
 				pd = d;
-				n = obj_image_color(image, bg_image[caching], d, &inithist);
+				n = obj_image_color(image, 
+						bgpixmap[bgindex].image, d, &inithist);
 				if (n == -1)
 					goto expand;
 			}
@@ -1659,7 +1674,7 @@ obj_image_trans(image, x, y)
 		fprintf(stderr, "  image=%dx%d+%d+%d",
 			image->width, image->height, x, y);
 		if (COMPLEX_BGIMAGE)
-			fprintf(stderr, "  bg_image[caching]=%dx%d", bw, bh);
+			fprintf(stderr, "  bgpixmap[bgindex].image=%dx%d", bw, bh);
 		fprintf(stderr, "\n");
 	}
 	image->trans = -1;	/* XXX: need recalculation to redraw? */
@@ -1678,12 +1693,12 @@ obj_draw_image(target, x, y, obj)
 	int private = mgp_flag & FL_PRIVATE;
 
 	image = obj->data.image.image;
-	if (obj->data.image.xzoom != 100 || obj->data.image.yzoom != 100) {
+	if (obj->data.image.xzoom != 100.0 || obj->data.image.yzoom != 100.0) {
 		timage = image;
 		image = zoom(image,
 			obj->data.image.xzoom, obj->data.image.yzoom, verbose);
 		if (!image) {
-			fprintf(stderr, "image zoom (%dx%d) failed in obj_draw_image\n",
+			fprintf(stderr, "image zoom (%0.2fx%0.2f) failed in obj_draw_image\n",
 				obj->data.image.xzoom, obj->data.image.yzoom);
 			exit(1);
 		}
@@ -1714,11 +1729,28 @@ obj_draw(state, target, xpos, ypos)
 {
 	struct render_object *obj;
 	int x = 0, y = 0;
-	int width, height;
+	int width, height, xwidth, xheight;
 	u_long fore;
 	u_int code;
 	char *registry;
 	XChar2b kch[2];
+#define MAXDRAWAREA 1024
+	struct {
+		int x, y, width, height;
+	} drawarea[MAXDRAWAREA];
+	int areaindex = 0;
+#define addarea(X) \
+{\
+	if (areaindex == MAXDRAWAREA){\
+		fprintf(stderr, "too many drawarea (increase MAXDRAWAREA)\n");\
+			exit(1);\
+	}\
+	drawarea[areaindex].x = x;\
+	drawarea[areaindex].y = y - obj->data.X->ascent;\
+	drawarea[areaindex].width = obj->data.X->charlen+1;\
+	drawarea[areaindex].height = obj->data.X->height+1;\
+	areaindex ++;\
+}
 #if 0
 	char ch[2];
 #endif
@@ -1746,12 +1778,18 @@ obj_draw(state, target, xpos, ypos)
 	bcolor = back_color[caching];
 	for (obj = state->obj; obj; obj = obj->next) {
 #ifdef VFLIB
-		if (obj->type == O_VFONT)
+		if (obj->type == O_VFONT){
+			xwidth = obj->data.vfc->width;
+			xheight = obj->data.vfc->height;
 			break;
+		}
 #endif /* VFLIB */
 #ifdef FREETYPE
-		if (obj->type == O_TFONT)
+		if (obj->type == O_TFONT){
+			xwidth = obj->data.tfc->width;
+			xheight = obj->data.tfc->height;
 			break;
+		}
 #endif /* FREETYPE */
 	}
 	if (obj != NULL) {	/* VFONT exist */
@@ -1767,7 +1805,7 @@ obj_draw(state, target, xpos, ypos)
 			int bpl;
 			XColor col;
 
-			bim = bg_ximageinfo[caching]->ximage;
+			bim = bgpixmap[bgindex].ximageinfo->ximage;
 			bw = bim->width;
 			bh = bim->height;
 			ox = xpos;
@@ -1776,22 +1814,22 @@ obj_draw(state, target, xpos, ypos)
 			by = oy % bh;
 			if (bw == 1) {
 				r = g = b = 0;
-				bpl = bg_image[caching]->pixlen;
-				bp = bg_image[caching]->data + by * bpl;
+				bpl = bgpixmap[bgindex].image->pixlen;
+				bp = bgpixmap[bgindex].image->data + by * bpl;
 				for (y = 0;
 				     y < height;
 				     y++, by++, bp += bpl) {
 					if (by == bh)
 						by = 0;
 					p = memToVal(bp, bpl);
-					if (TRUEP(bg_image[caching])) {
+					if (TRUEP(bgpixmap[bgindex].image)) {
 						r += TRUE_RED(p) << 8;
 						g += TRUE_GREEN(p) << 8;
 						b += TRUE_BLUE(p) << 8;
 					} else {
-						r += bg_image[caching]->rgb.red[p];
-						g += bg_image[caching]->rgb.green[p];
-						b += bg_image[caching]->rgb.blue[p];
+						r += bgpixmap[bgindex].image->rgb.red[p];
+						g += bgpixmap[bgindex].image->rgb.green[p];
+						b += bgpixmap[bgindex].image->rgb.blue[p];
 					}
 					p = XGetPixel(bim, 0, by);
 					for (x = 0; x < width; x++)
@@ -1810,7 +1848,7 @@ obj_draw(state, target, xpos, ypos)
 #if 0
 			fprintf(stderr, "bim=%dx%d, r=%x, g=%x, b=%x, "
 				"bcolor=%x\n",
-				bg_image[caching]->width, bg_image[caching]->height,
+				bgpixmap[bgindex].image->width, bgpixmap[bgindex].image->height,
 				col.red, col.green, col.blue, bcolor);
 #endif
 			} else {
@@ -1851,17 +1889,27 @@ obj_draw(state, target, xpos, ypos)
 			if (obj->type == O_VFONT) {
 				(void)vfc_image(obj->data.vfc,
 					obj->fore, bcolor, xim, x, y);
+				addarea(vfc);
 			}
 #endif /* VFLIB */
 #ifdef FREETYPE
 			if (obj->type == O_TFONT) {
 				(void)tfc_image(obj->data.tfc,
 					obj->fore, bcolor, xim, x, y);
+				addarea(tfc);
 			}
 #endif /* FREETYPE */
 		}
+#if 0
 		XPutImage(display, target, gcfore, xim, 0, 0,
 			xpos, ypos, width, height);
+#else
+		for (i = 0; i < areaindex; i ++)
+			XPutImage(display, target, gcfore, xim,
+				drawarea[i].x, drawarea[i].y,
+				drawarea[i].x + xpos, ypos + drawarea[i].y,
+				drawarea[i].width, drawarea[i].height);
+#endif
 		XDestroyImage(xim);
 		if (mgp_flag & FL_GLYPHEDGE) {
 			XDrawLine(display, target, gcfore, 0, ypos,
@@ -1877,6 +1925,7 @@ obj_draw(state, target, xpos, ypos)
 				0, ypos + height,
 				state->width - 1, ypos + height);
 		}
+
 	}
 #endif /* RASTERLIB */
 	for (obj = state->obj; obj; obj = obj->next) {
@@ -2168,7 +2217,7 @@ x_setfont(xfont, csize, registry, truescalable)
 
 	if (verbose) {
 		fprintf(stderr, "need font <%s:%d:%s>\n",
-			xfont, csize, registry);
+			xfont, csize, registry?registry:"NULL");
 	}
 
 	/*
@@ -2369,7 +2418,7 @@ draw_onechar_x(state, code, x, y, size, registry, lastchar)
 			fprintf(stderr, "X11 font %s:%d:%s has bogus "
 				"font metric for glyph 0x%04x\n"
 				"\tcs->width=%d, source=%s, coffset=0x%04x\n",
-				seed, char_size, registry,
+				seed, char_size, registry?registry:"NULL",
 				code, cs->width, metricsource, coffset);
 		}
 		cs = &xfontstruct->max_bounds;
@@ -2416,7 +2465,7 @@ back_gradation(state, cg0)
 	int srcwidth, srcheight;
 	int dstwidth, dstheight;
 	int dir, numcolor;
-	int xzoomrate, yzoomrate;
+	float xzoomrate, yzoomrate;
 	int hquality, vquality;
 
 	Image *myimage, *image;
@@ -2494,12 +2543,13 @@ back_gradation(state, cg0)
 	}
 #endif
 
-	xzoomrate = 100 * dstwidth / srcwidth;
-	yzoomrate = 100 * dstheight / srcheight;
+	xzoomrate = 100.0 * dstwidth / srcwidth;
+	yzoomrate = 100.0 * dstheight / srcheight;
 
 	/* performace enhance hack for special case */
 	if (dir % 90 == 0) {
-		int *p, *q, *r;
+		float *q;
+		int *p, *r;
 
 		/*
 		 * 0 or 180: reduce width
@@ -2510,14 +2560,19 @@ back_gradation(state, cg0)
 		r = (dir % 180 == 0) ? &dstwidth : &dstheight;
 
 		/* rely upon use X11 background image tiling. */
+		*q = (float) 100.0;
+#ifndef DITHERED_BGRAD
 		*p = 1;
-		*q = 100;
 		*r = 1;
+#else
+		*p = 3;
+		*r = 3;
+#endif
 	}
 
 	if (verbose) {
 		fprintf(stderr, "raw: %d,%d qu: %d,%d "
-			"dst: %d,%d src: %d,%d zoom: %d,%d\n",
+			"dst: %d,%d src: %d,%d zoom: %0.2f,%0.2f\n",
 			cg->ct_width, cg->ct_height,
 			hquality, vquality,
 			dstwidth, dstheight, srcwidth, srcheight,
@@ -2534,25 +2589,27 @@ back_gradation(state, cg0)
 		myimage = reduce(myimage, numcolor, verbose);
 
 	if (verbose) {
-		fprintf(stderr, "background zoomrate: (%d,%d)\n",
+		fprintf(stderr, "background zoomrate: (%0.2f,%0.2f)\n",
 			xzoomrate, yzoomrate);
 		fprintf(stderr, "background zoom mode %d: "
 			"(%d, %d)->(%d, %d)[%d]\n", cg->ct_zoomflag,
 			srcwidth, srcheight, dstwidth, dstheight, b_quality);
 	}
 
-	if (xzoomrate != 100 || yzoomrate != 100) {
+	if (xzoomrate != 100.0 || yzoomrate != 100.0) {
 		image = myimage;
 		myimage = zoom(image, xzoomrate, yzoomrate, verbose);
 		if (!image) {
-			fprintf(stderr, "image zoom (%dx%d) failed in back_gradataion\n",
+			fprintf(stderr, "image zoom (%0.2fx%0.2f) failed in back_gradataion\n",
 				xzoomrate, yzoomrate);
 			exit(1);
 		}
 		freeImage(image);
 	}
 
+#ifndef COLOR_BUGFIX
 	if (private) free_alloc_colors(&back_clr);
+#endif
 	ximageinfo = imageToXImage(display, screen, visual, depth, myimage,
 		private, 0, 1, verbose);
 	if (!ximageinfo) {
@@ -2560,29 +2617,7 @@ back_gradation(state, cg0)
 		cleanup(-1);
 	}
 
-	mypixmap = ximageToPixmap(display, DefaultRootWindow(display),
-		ximageinfo);
-	if (mypixmap == None) {
-		fprintf(stderr, "Cannot create image in server\n");
-		cleanup(-1);
-	}
-
-	/* draw background */
-	if (!caching) 
-		XSetWindowBackgroundPixmap(display, state->target, mypixmap);
-	else  
-		XSetPixmapBackgroundPixmap(display, state->target, 
-			mypixmap, ximageinfo->ximage->width, ximageinfo->ximage->height);
-
-	/* finish */
-	XFreePixmap(display, mypixmap);
-
-	if (bg_image[caching]) {
-		freeXImage(bg_image[caching], bg_ximageinfo[caching]);
-		freeImage(bg_image[caching]);
-	}
-	bg_ximageinfo[caching] = ximageinfo;
-	bg_image[caching] = myimage;
+	regist_background_pixmap(ximageinfo, myimage);
 
 	XUndefineCursor(display, window);
 	XFlush(display);
@@ -2604,7 +2639,7 @@ image_load(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag
 	XImageInfo *ximageinfo;
 	u_int image_posx;
 	int width, height;
-	int xzoomrate, yzoomrate;
+	float xzoomrate, yzoomrate;
 	int	private = mgp_flag & FL_PRIVATE;
 	static Cursor curs;
 	static char backfile[MAXPATHLEN];
@@ -2624,7 +2659,7 @@ image_load(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag
 	width = myimage->width;
 	height = myimage->height;
 
-	if (myimage->depth == 1) {
+	if (myimage->depth == 1 && myimage->trans < 0) {
 		XColor xc;
 
 		xc.flags = DoRed | DoGreen | DoBlue;
@@ -2663,17 +2698,19 @@ image_load(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag
 		else
 			xzoomrate = ximagesize;
 	} else
-		xzoomrate = 100;
+		xzoomrate = 100.0;
 	if (yimagesize != 0) {
 		if (!zoomflag)
 			yzoomrate = state->height * yimagesize / height;
 		else
 			yzoomrate = yimagesize;
 	} else
-		yzoomrate = 100;
+		yzoomrate = 100.0;
 #else
-	xzoomrate = ximagesize;
-	yzoomrate = yimagesize;
+	if (!ximagesize) ximagesize = 100;
+	if (!yimagesize) yimagesize = 100;
+	xzoomrate = (float) ximagesize;
+	yzoomrate = (float) yimagesize;
 	image_zoomratio(state, &xzoomrate, &yzoomrate, zoomflag, width, height);
 #endif
 
@@ -2689,42 +2726,16 @@ image_load(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag
 			freeImage(image);
 		}
 
+#ifndef COLOR_BUGFIX
 		if (private) free_alloc_colors(&back_clr);
+#endif
 		ximageinfo= imageToXImage(display, screen, visual, depth,
 				myimage, private, 0, 1, verbose);
 		if (ximageinfo == NULL) {
 			fprintf(stderr, "Cannot convert Image to XImage\n");
 			cleanup(-1);
 		}
-
-		mypixmap = ximageToPixmap(display,
-				RootWindow(display, screen), ximageinfo);
-		if (mypixmap == None) {
-			fprintf(stderr, "Cannot create image in server\n");
-			cleanup(-1);
-		}
-
-		if (!caching)
-			XSetWindowBackgroundPixmap(display, state->target, mypixmap);
-		else  {
-			XSetPixmapBackgroundPixmap(display, state->target, mypixmap,
-				ximageinfo->ximage->width, ximageinfo->ximage->height);
-		}
-
-		XFreePixmap(display, mypixmap);
-		strcpy(backfile, filename);
-		backnumcolor = numcolor;
-		backzoom = zoomflag;
-		backx = ximagesize;
-		backy = yimagesize;
-
-		if (bg_image[caching]) {
-			freeXImage(bg_image[caching], bg_ximageinfo[caching]);
-			freeImage(bg_image[caching]);
-		}
-		bg_ximageinfo[caching] = ximageinfo;
-		bg_image[caching] = myimage;
-
+		regist_background_pixmap(ximageinfo, myimage);
 		goto end;
 	}
 
@@ -2761,7 +2772,7 @@ image_load_ps(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomf
 	char fullname[MAXPATHLEN];
 	char *imagefile;
 	int width, height;
-	int xzoom, yzoom;
+	float xzoom, yzoom;
 	char *p;
 
 	/* wait for a while, please. */
@@ -2798,23 +2809,23 @@ image_load_ps(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomf
 	}
 	if (ximagesize != 0) {
 		if (!zoomflag)
-			xzoom = state->width * ximagesize / width;
+			xzoom = ((float) state->width * ximagesize) / width;
 		else
-			xzoom = ximagesize;
+			xzoom = (float) ximagesize;
 		width = width * xzoom / 100;
 	} else
-		xzoom = 100;
+		xzoom = 100.0;
 	if (yimagesize != 0) {
 		if (!zoomflag)
-			yzoom = state->height * yimagesize / height;
+			yzoom = ((float) state->height * yimagesize) / height;
 		else
 			yzoom = yimagesize;
 		height = height * yzoom / 100;
 	} else
-		yzoom = 100;
+		yzoom = 100.0;
 #else
-	xzoom = ximagesize;
-	yzoom = yimagesize;
+	xzoom = (float) ximagesize;
+	yzoom = (float) yimagesize;
 	image_zoomratio(state, &xzoom, &yzoom, zoomflag, width, height);
 	width = width * xzoom / 100;
 	height = height * yzoom / 100;
@@ -3033,6 +3044,7 @@ draw_bar(state, cp)
 	XFillRectangle(display, state->target, gcsbar, st, state->ypos + width, len, swidth);
 
 	state->ypos += width + swidth + VERT_GAP(char_size[caching]) / 2;
+	state->maxascent = width + swidth;
 }
 
 static void
@@ -3244,13 +3256,6 @@ fail:
 		argv[cp->cta_argc+1] = geom_arg2;
 	}
 
-#if 0
-	xloc = state->linewidth;
-#endif
-	state->linewidth = xsiz;
-	xloc = set_position(state);
-	yloc = state->ypos;
-
 	if ((window_id = checkchildwin(cp)) != (Window)-1)
 		goto finish;	/*already running*/
 
@@ -3276,7 +3281,7 @@ fail:
 		regchild(pid, cp, window_id, cp->cta_flag);
 
 	if (window_id != -1)
-		reparent_child_window(window_id, xloc, yloc);
+		reparent_child_window(window_id, window_width, window_height);
 	else {
 		if (mgp_flag & FL_VERBOSE) {
 			fprintf(stderr, "WARN: %%xsystem can not find child window:");
@@ -3289,15 +3294,19 @@ fail:
 		return;
 	}
 
-#if 0
-	state->linewidth += xsiz;
-	state->maxascent += ysiz;
-#endif
-
 finish:
 	XGetGeometry(display, window_id, &dumwin, 
 		&xloc, &yloc, &xsiz, &ysiz, &dumint, &dumint);
+	state->linewidth = xsiz;
+	xloc = set_position(state);
+	yloc = state->ypos;
+	XMoveWindow(display, window_id, xloc, yloc);
 	state->ypos += ysiz;
+
+#if 0 /* not implemented yet */
+	state->linewidth += xsiz;
+	state->maxascent += ysiz;
+#endif
 }
 
 
@@ -3369,7 +3378,8 @@ draw_reinit(state)
 	struct render_state *state;
 {
 	/* invalidate the background image cache */
-	bg_image_prev[0] = NULL;
+
+	bg_ctl_last = NULL;
 	x_registerseed(state, NULL, NULL);
 }
 
@@ -3638,20 +3648,21 @@ image_setcolor(state)
 			Pixel d;
 
 			/* XXX: use background color of center position */
-			x = (obj->x + image->width/2) % bg_image[caching]->width;
-			y = (state->ypos + image->height/2) % bg_image[caching]->height;
-			bpl = bg_image[caching]->pixlen;
-			p = bg_image[caching]->data 
-				+ (bg_image[caching]->width * y + x) * bpl;
+			x = (obj->x + image->width/2) % bgpixmap[bgindex].image->width;
+			y = (state->ypos + image->height/2) 
+					% bgpixmap[bgindex].image->height;
+			bpl = bgpixmap[bgindex].image->pixlen;
+			p = bgpixmap[bgindex].image->data 
+				+ (bgpixmap[bgindex].image->width * y + x) * bpl;
 			d = memToVal(p, bpl);
-			if (bg_image[caching]->type == ITRUE) {
+			if (bgpixmap[bgindex].image->type == ITRUE) {
 				back.red   = TRUE_RED(d) << 8;
 				back.green = TRUE_GREEN(d) << 8;
 				back.blue  = TRUE_BLUE(d) << 8;
 			} else {
-				back.red   = bg_image[caching]->rgb.red  [d];
-				back.green = bg_image[caching]->rgb.green[d];
-				back.blue  = bg_image[caching]->rgb.blue [d];
+				back.red   = bgpixmap[bgindex].image->rgb.red  [d];
+				back.green = bgpixmap[bgindex].image->rgb.green[d];
+				back.blue  = bgpixmap[bgindex].image->rgb.blue [d];
 			}
 		}
 		for (i = 0; i < image->rgb.used; i++) {
@@ -3791,6 +3802,7 @@ x_registerseed(state, seed, registry)
 
 	cp = NULL;
 	for (cp = state->xfont; cp; cp = cp->ct_next) {
+		if (!cp->ctc2_value2) continue;
 		if (strcmp(cp->ctc2_value2, registry) == 0)
 			break;
 	}
@@ -3823,42 +3835,6 @@ x_findseed(state, registry)
 	return "*-*-*";		/*anything, canonicalized*/
 }
 
-static void 
-XSetPixmapBackground(display, target, bgcolor)
-	Display *display;
-	Drawable target;
-	u_long	bgcolor;
-{
-	XSetForeground(display, gc_cache, bgcolor);
-	XFillRectangle(display, cacheback,
-		gc_cache, 0, 0, window_width, window_height);
-}
-
-static void 
-XSetPixmapBackgroundPixmap(display, target, pixmap, width, height)
-	Display *display;
-	Drawable target;
-	Pixmap	pixmap;
-	int	width, height;
-{
-	int	x, y;
-
-	for (y = 0; y < window_height; y += height)
-		for (x = 0; x < window_width; x += width){
-			XCopyArea(display, pixmap, cacheback, gc_cache, 0, 0,
-				width, height, x, y);
-		}
-}
-
-static void 
-XClearPixmap(display, target)
-	Display *display;
-	Drawable target;
-{
-	XCopyArea(display, cacheback, cachewin, gc_cache, 
-			0, 0, window_width, window_height, 0, 0);
-}
-
 /* cache specified page */
 static void
 cache_page(state, page)
@@ -3892,6 +3868,8 @@ static void
 set_from_cache(state)
 	struct render_state *state;
 {
+	int	i;
+
 	char_size[0] = char_size[1];
 	horiz_gap[0] = horiz_gap[1];
 	vert_gap[0] = vert_gap[1];
@@ -3900,24 +3878,13 @@ set_from_cache(state)
 	ctrl_color[0] = ctrl_color[1];
 	b_quality[0] = b_quality[1];
 
-	bg_image_ctl[0] = bg_image_ctl[1];
-	bg_image_prev[0] = bg_image_prev[1];
-	bg_image[0] = bg_image[1];
-	bg_ximageinfo[0] = bg_ximageinfo[1];
-
-	bg_image_ctl[1] = NULL;
-	bg_image_prev[1] = NULL;
-	bg_image[1] = NULL;
-	bg_ximageinfo[1] = NULL;
-
 	memcpy(state, &cache_state, sizeof(struct render_state));	
 	state->target = window;
 
 	XSetForeground(display, gcfore, fore_color[0]);
 	XSetBackground(display, gcfore, back_color[0]);
-	XCopyArea(display, cacheback, cachetmpback, gc_cache,
-                  0, 0,  window_width, window_height, 0, 0);
-	XSetWindowBackgroundPixmap(display, state->target, cachetmpback);
+	bg_ctl = bg_ctl_last = bg_ctl_cache;
+	set_background_pixmap(bg_ctl);
 
 	switch(cache_effect){
 		case 1:
@@ -3935,12 +3902,22 @@ set_from_cache(state)
 }
 
 void
-reset_cache_bgimage()
+reset_background_pixmap()
 {
-	bg_image_ctl[1] = NULL;
-	bg_image_prev[1] = NULL;
-	bg_image[1] = NULL;
-	bg_ximageinfo[1] = NULL;
+	int	i = 0;
+
+	bg_ctl_last = NULL;
+
+	for (i = 0; i < MAXBGPIXMAP; i ++) {
+		if (bgpixmap[i].image){
+			XFreePixmap(display, bgpixmap[i].pixmap); 
+			freeXImage(bgpixmap[i].image, bgpixmap[i].ximageinfo);
+			freeImage(bgpixmap[i].image);
+		}
+		bgpixmap[i].ctl = NULL;
+		bgpixmap[i].image = NULL;
+		bgpixmap[i].ximageinfo = NULL;
+	}
 }
 
 static void
@@ -4004,3 +3981,166 @@ int	page;
 	cache_value  =  pcache.value;
 	pcache.flag  = 0;
 }
+
+
+/*
+	predraw: if this page contains texts only, 
+			   draw page in pixmap once, then copy to window.
+*/
+static void
+predraw(state)
+	struct render_state *state;
+{
+	if (!caching && cached_page != state->page 
+			&& page_attribute[state->page].pg_text){
+		cache_page(&cache_state, state->page);
+		set_from_cache(state);
+		pcache_process(state->page);
+	}
+}
+
+
+static void
+get_background_pixmap(ctl, state)
+	struct ctrl *ctl;
+	struct render_state *state;
+{
+	int	i;
+
+	/*
+	 * check if background is already cached
+	 */
+	for (i = 0; i < MAXBGPIXMAP; i ++){
+		if (bgpixmap[i].ctl && ctlcmp(ctl, bgpixmap[i].ctl) == 0){
+			bgindex = i;
+			return;			
+		}
+	}
+
+	if (i == MAXBGPIXMAP){
+		/* this background is not cached, we have to generate one */
+		switch(ctl->ct_op){
+		case CTL_BIMAGE: 
+			image_load(state, ctl->ctm_fname, ctl->ctm_numcolor,
+						ctl->ctm_ximagesize, ctl->ctm_yimagesize, 1,
+						ctl->ctm_zoomflag, 0);
+			break;
+		case CTL_BGRAD:
+			back_gradation(state, &ctl->ct_val.ctrl_grad);
+			break;
+		case CTL_BACK:
+		default:
+			break;
+		}
+	}
+}
+
+static void
+regist_background_pixmap(ximageinfo, image)
+	XImageInfo	*ximageinfo;
+	Image		*image;
+{
+	Pixmap	pixmap;
+	int	i, j;
+
+	/* search empty slot */	
+	for (i = 0; i < MAXBGPIXMAP; i ++){
+		if (bgpixmap[i].ctl == NULL)
+			break;
+	}
+
+	if (i == MAXBGPIXMAP){
+		/* no empty slot, we need to make one  */
+		XFreePixmap(display, bgpixmap[MAXBGPIXMAP -1].pixmap); 
+		freeXImage(bgpixmap[MAXBGPIXMAP -1].image, 
+					bgpixmap[MAXBGPIXMAP -1].ximageinfo);
+		freeImage(bgpixmap[MAXBGPIXMAP -1].image);
+		for (j = MAXBGPIXMAP -2; j >= 0; j --){
+			bgpixmap[j +1].ctl = bgpixmap[j].ctl;
+			bgpixmap[j +1].pixmap = bgpixmap[j].pixmap;
+			bgpixmap[j +1].image = bgpixmap[j].image;
+			bgpixmap[j +1].ximageinfo = bgpixmap[j].ximageinfo;
+		}
+		i = 0;
+	}
+
+	pixmap = ximageToPixmap(display, 
+			RootWindow(display, screen), ximageinfo);
+	bgpixmap[i].ctl = bg_ctl;
+	bgpixmap[i].pixmap = pixmap;
+	bgpixmap[i].image = image;
+	bgpixmap[i].ximageinfo = ximageinfo;
+	bgindex = i;
+}
+
+static void
+set_background_pixmap(ctl)
+	struct ctrl *ctl;
+{
+	int	i;
+
+	switch(ctl->ct_op){
+	case CTL_BIMAGE: 
+	case CTL_BGRAD:
+		for (i = 0; i < MAXBGPIXMAP; i ++){
+			if (bgpixmap[i].ctl && ctlcmp(ctl, bgpixmap[i].ctl) == 0)
+				break;	
+		}
+		if (i == MAXBGPIXMAP){
+			fprintf(stderr, "fatal error in set_background_image()\n");
+			cleanup(-1);
+		}
+		XSetWindowBackgroundPixmap(display, window, bgpixmap[i].pixmap);
+		break;
+	case CTL_BACK:
+		XSetWindowBackground(display, window, ctl->ctl_value);
+		break;
+	default:
+		fprintf(stderr, "fatal error in set_background_image()\n");
+		cleanup(-1);
+		break;
+	}
+}
+
+/*
+ * Clear target pixmap 
+ */
+static void
+XClearPixmap(display, target)
+	Display *display;
+	Drawable target;
+{
+	int	i; 
+	int x, y, width, height;
+	XImage *xim;
+
+	switch(bg_ctl->ct_op){
+	case CTL_BIMAGE: 
+	case CTL_BGRAD:
+		for (i = 0; i < MAXBGPIXMAP; i ++){
+			if (bgpixmap[i].ctl && ctlcmp(bg_ctl, bgpixmap[i].ctl) == 0)
+				break;	
+		}
+		if (i == MAXBGPIXMAP){
+			fprintf(stderr, "fatal error in set_background_image()\n");
+			cleanup(-1);
+		}
+
+		xim = bgpixmap[i].ximageinfo->ximage;
+		for (y = 0; y < window_height; y += xim->height)
+			for (x = 0; x < window_width; x += xim->width)
+				XPutImage(display, target, gc_cache, 
+					xim, 0, 0, x, y, 
+					xim->width, xim->height);
+		break;
+	case CTL_BACK:
+		XSetForeground(display, gc_cache, bg_ctl->ctl_value);
+		XFillRectangle(display, target,
+			gc_cache, 0, 0, window_width, window_height);
+		break;
+	default:
+		fprintf(stderr, "fatal error in set_background_image()\n");
+		cleanup(-1);
+		break;
+	}
+}
