@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: mgp.h,v 1.108 1999/06/06 17:55:18 nishida Exp $
+ * $Id: mgp.h,v 1.118 2000/03/05 07:26:36 nishida Exp $
  */
 
 #include <stdio.h>
@@ -101,7 +101,7 @@
 #define PAGELIST_KFONT	"k14"
 
 #define DEFAULT_CHARSIZE	10	/* 10% of height */
-#define DEFAULT_HGAP		10
+#define DEFAULT_HGAP		0
 #define DEFAULT_VGAP		15
 #define DEFAULT_BQUALITY	100
 
@@ -120,7 +120,7 @@
 #ifndef UUDECODE
 #define	UUDECODE	"/usr/bin/uudecode"
 #endif
-#ifndef UUDECODE
+#ifndef GUNZIP
 #define	GUNZIP		"/usr/contrib/bin/gunzip"
 #endif
 
@@ -133,6 +133,8 @@
 #define MAXSTYLE	100
 #define MAXFONTDEF	100
 
+#define MAXBGPIXMAP	2
+
 #define SP_NONE		0
 #define SP_SHRINK	1
 #define SP_LCUTIN	2
@@ -327,6 +329,7 @@ struct ctl_words {
 #define T_STR2	'C'
 #define T_INT	'i'
 #define T_LONG	'l'
+#define T_DOUBLE 'f'
 #define T_SP	'x'
 #define T_VOID	'-'
         char *ctl_string;
@@ -340,6 +343,7 @@ struct page_attribute {
 	u_int pg_linenum;
 	u_int pg_b_numcolor;	/* background gradation number of colors */
 	u_int pg_b_dir;		/* background gradation deg */
+	u_int pg_text;		/* this page is text only */
 };
 extern struct page_attribute page_attribute[MAXPAGE];
 
@@ -424,7 +428,7 @@ struct render_object {
 		} xfont;
 		struct {
 			Image *image;
-			int xzoom, yzoom;
+			float xzoom, yzoom;
 		} image;
 		struct {
 			u_int itype;
@@ -451,7 +455,6 @@ struct vfont {
 	int descent;
 	u_int xoff;
 	char *fontname;
-	XImage *image;
 	u_char *dbitmap;
 	u_int ref;
 };
@@ -473,7 +476,6 @@ struct tfont {
 	int descent;		/* (origin y axis) - (bottom of bitmap) */
 	int xoff;		/* (left of bitmap) - (origin x axis) */
 	char *fontname;
-	XImage *image;
 	u_char *dbitmap;
 	u_int ref;
 };
@@ -484,6 +486,13 @@ struct alloc_color {
 	long	*colors;
 };
 
+struct bgpixmap {
+	struct ctrl *ctl;
+	Pixmap	pixmap;
+	Image *image;
+	XImageInfo *ximageinfo;
+};
+
 /*
  * The following variables are defined in global.c, and therefore
  * they are available in both "mgp" and "mgp2ps" binary.
@@ -513,8 +522,8 @@ extern int window_width;
 extern int window_height;
 extern Pixmap pixmap;
 extern Pixmap cachewin;
-extern Pixmap cacheback;
-extern Pixmap cachetmpback;
+extern Pixmap cachetmp;
+extern struct bgpixmap bgpixmap[MAXBGPIXMAP];
 extern Colormap colormap;
 extern struct alloc_color image_clr;
 extern struct alloc_color back_clr;
@@ -582,6 +591,7 @@ extern char *vfcap_name;
 #ifdef FREETYPE
 extern char *freetypefontdir;
 extern char *freetypefont0;	/* font name to be used as a last resort */
+extern char *freetypemfont0;	/* font name to be used as a last resort */
 extern int unicode_map[65536];
 #endif
 #ifdef TTY_KEYINPUT
@@ -589,6 +599,7 @@ extern volatile int ttykey_enable;
 #endif
 extern char *back_clname;
 extern char *gsdevice;
+extern char *htmlimage;
 
 extern char *mgpwdir;
 extern char mgpwdirname[];
@@ -665,7 +676,7 @@ extern Image *make_XImage __P((byte *, unsigned int, unsigned int));
 
 /* postscript.c */
 extern int ps_boundingbox __P((char *, int *, int *, int *, int *));
-extern void image_zoomratio __P((struct render_state *, int *, int *, int, int, int));
+extern void image_zoomratio __P((struct render_state *, float *, float *, int, int, int));
 
 /* tfont.c */
 extern int tfc_setsize __P((u_int));
@@ -695,3 +706,8 @@ extern void usleep __P((u_int));
 #ifdef VFLIB
 #define RASTERLIB
 #endif
+
+#if 1
+#define DITHERED_BGRAD
+#define COLOR_BUGFIX
+#endif
