@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: print.c,v 1.81 1999/05/25 09:44:30 nishida Exp $
+ * $Id: print.c,v 1.86 1999/11/28 14:35:09 nishida Exp $
  */
 /*
  * Paper size selection code is based on psutil.c by Angus J. C. Duggan
@@ -54,6 +54,10 @@ static int painticon = 0;
 static u_int curlinenum;
 static u_int curpagenum;
 
+#ifdef FREETYPE
+static u_int usetfont = 0;
+#endif
+
 static struct imagepool {
 	struct ctrl *image;
 	int xoffset;
@@ -84,6 +88,7 @@ char *vfcap_name = NULL;	/*dummy*/
 #ifdef FREETYPE
 char *freetypefontdir = NULL;	/*dummy*/
 char *freetypefont0 = NULL;	/*dummy*/
+char *freetypemfont0 = NULL;	/*dummy*/
 #endif
 
 /* lang code */
@@ -250,8 +255,9 @@ static void print_8_image __P((XImage *));
 static void print_32_image __P((XImage *, Visual *));
 static void print_eps __P((FILE *, char *, int, int, int, int, int));
 static struct fontmap *findfont __P((int, int, char *));
+static struct fontmap *findttfont __P((int, int, char *));
 static char *fontstring __P((struct fontmap *));
-static void loadfont __P((struct fontmap *));
+static void loadfont __P((struct fontmap *, char *));
 
 /* image/ *.[ch] */
 extern char *Paths[];
@@ -294,8 +300,9 @@ print_out()
 		fp = stdout;
 
 	print_init();
-	for (i = 1; i <= maxpage; i ++)
+	for (i = 1; i <= maxpage; i ++) {
 		scan_font(i);
+	}
 	for (i = 1; i <= maxpage; i ++)
 		print_page(i);
 
@@ -373,9 +380,24 @@ scan_font(page)
 
 			case CTL_TEXT:
 				if (strstr(cp->ctc_value, "\033$B")
-				 || strstr(cp->ctc_value, "\033$@"))
-					loadfont(curfont[KANJI]);
-				loadfont(curfont[ASCII]);
+				 || strstr(cp->ctc_value, "\033$@")) {
+					if (curfont[KANJI]) {
+						loadfont(curfont[KANJI],
+							curfont[KANJI]->font);
+					} else {
+						fprintf(stderr,
+"unable to find proper font for %s\n", "kanji");
+						exit(1);
+					}
+				}
+				if (curfont[ASCII]) {
+					loadfont(curfont[ASCII],
+						curfont[ASCII]->font);
+				} else {
+					fprintf(stderr,
+"unable to find proper font for %s\n", "ascii");
+					exit(1);
+				}
 				break;
 			default:
 				break;
@@ -431,6 +453,12 @@ print_init()
 	fprintf(fp, "/XBODY WIDTH XMARGIN 2 mul sub def\n");
 	fprintf(fp, "/vertgap %d def /horizgap %d def\n", vert_gap[0], horiz_gap[0]);
 
+	fprintf(fp, "/resety {/ymin 0 def /ymax 0 def} def\n");
+	fprintf(fp, "/setymax {dup ymax gt {dup /ymax exch def} if pop} def\n");
+	fprintf(fp, "/setymin {dup ymin lt {dup /ymin exch def} if pop} def\n");
+	fprintf(fp, "/calcy {false charpath flattenpath pathbbox setymax pop setymin pop} def\n");
+	fprintf(fp, "resety\n");
+
 	/* define writebox */
 	fprintf(fp, "/writebox {\n");
 	fprintf(fp, "  XMARGIN YMARGIN -1 mul moveto "
@@ -450,14 +478,16 @@ print_init()
 	/* ypos = ypos - (charsize * (1 + vert_gap / 100)) */
 	fprintf(fp, "/NL {\n");
 	fprintf(fp, "  charsize imgsize gt\n");
-	fprintf(fp, "    { vertgap 100 div 1 add charsize mul }\n");
-	fprintf(fp, "    { vertgap 100 div charsize mul imgsize add }\n");
+	fprintf(fp, "	ymax ymin sub /csize exch def\n");
+	fprintf(fp, "	csize 0 eq {/csize charsize def} if\n");
+	fprintf(fp, "    { vertgap 100 div 1 add csize mul }\n");
+	fprintf(fp, "    { vertgap 100 div csize mul imgsize add }\n");
 	fprintf(fp, "  ifelse\n");
 	fprintf(fp, "  ypos exch sub /ypos exch def\n");
 	fprintf(fp, "} bind def\n");
 
 	/* show with line wrapping */
-	fprintf(fp, "/initcharsize { /charsize 0 def /imgsize 0 def } def\n");
+	fprintf(fp, "/initcharsize { /charsize 0 def /imgsize 0 def resety} def\n");
 	fprintf(fp, "initcharsize\n");
 	fprintf(fp, "/setcharsize {\n");
 	fprintf(fp, "  dup charsize gt { dup /charsize exch def } if pop\n");
@@ -733,8 +763,12 @@ line_start()
 	nimagepool = 0;
 	ntextpool = 0;
 
-	if (curlinenum == 0)
+	if (curlinenum == 0){
 		print_pagebackground();
+#if 0
+		fprintf(fp, "initcharsize %d setcharsize\n", char_size[0]);
+#endif
+	}
 }
 
 static void
@@ -769,14 +803,17 @@ line_end()
 	/*
 	 * if we have nothing on the line, skip it
 	 */
-	if (!ntextpool && !nimagepool)
+	if (!ntextpool && !nimagepool){
+		fprintf(fp, "initcharsize %d setcharsize\n", char_size[0]);
 		goto done;
+	}
 
 	/*
 	 * push strings in reverse order.
 	 */
 	fprintf(fp, "initcharsize\n");
 	fprintf(fp, "0 %% sentinel for text width computation\n");
+	fprintf(fp, "gsave newpath 0 0 moveto\n");
 
 	for (i = ntextpool - 1; 0 <= i; i--) {
 		if (!textpool[i].text) {
@@ -784,14 +821,22 @@ line_end()
 			continue;
 		}
 #if 0
-		loadfont(textpool[i].font);
+		loadfont(textpool[i].font, textpool[i].charset);
 #endif
+
+		fprintf(fp, "%d setcharsize %d %s %s calcy\n",
+			textpool[i].size, textpool[i].size,
+			fontstring(textpool[i].font),
+			textpool[i].text);
+
 		fprintf(fp, "%d setcharsize %d %s %s "
 			"1 copy stringwidth pop 3 2 roll add\n",
 			textpool[i].size, textpool[i].size,
 			fontstring(textpool[i].font),
 			textpool[i].text);
 	}
+	fprintf(fp, "grestore\n");
+
 	if (mgp_flag & FL_VERBOSE) {
 		fprintf(fp, "%% stack should have: str3 str2 str1 width\n");
 		fprintf(fp, "%% alignment: %s\n", alignstr);
@@ -812,12 +857,12 @@ line_end()
 	fprintf(fp, "/totlen exch def\n");
 	fprintf(fp, "/inmargin true def /fillzero XMARGIN def\n");
 	fprintf(fp, "%sdefxpos ", alignstr);
-	fprintf(fp, "xpos ypos charsize sub moveto\n");
+	fprintf(fp, "xpos ypos ymax sub moveto\n");
 
 	for (i = 0; i < ntextpool; i++) {
 		if (textpool[i].text) {
 #if 0
-			loadfont(textpool[i].font);
+			loadfont(textpool[i].font, textpool[i].charset);
 #endif
 			fprintf(fp, "%d %s 1 copy stringwidth pop ",
 				textpool[i].size,
@@ -1256,7 +1301,7 @@ image_remember(cp, pool)
 {
 	char *epsname;
 	Image *myimage;
-	int xzoom, yzoom;
+	float xzoom, yzoom;
 	struct render_state state;
 
 	if ((epsname = checkeps(cp->ctm_fname))) {
@@ -1286,14 +1331,14 @@ image_remember(cp, pool)
 			break;
 		}
 #else
-		xzoom = xim;
-		yzoom = yim;
+		xzoom = (float) xim;
+		yzoom = (float) yim;
 		state.width = window_width;
 		state.height = window_height;
 		image_zoomratio(&state, &xzoom, &yzoom, cp->ctm_zoomflag,
 			width, height);
-		swidth = width * xzoom / 100;
-		sheight = height * yzoom / 100;
+		swidth = (int) width * xzoom / 100;
+		sheight = (int) height * yzoom / 100;
 #endif
 		pool->xsiz = swidth;
 		pool->ysiz = sheight;
@@ -1329,7 +1374,7 @@ image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 	XImageInfo *ximageinfo;
 	XImage	*print_image;
 	int width, height;
-	int xzoomrate, yzoomrate;
+	float xzoomrate, yzoomrate;
 	static Cursor curs;
 	u_int	print_width, print_height;
 	struct render_state state;
@@ -1370,20 +1415,20 @@ image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 #if 0
 	if (ximagesize != 0 && yimagesize != 0) {
 		if (!zoomflag) {
-			xzoomrate = window_width * ximagesize / width;
-			yzoomrate = window_height * yimagesize / height;
+			xzoomrate = (float) window_width * ximagesize / width;
+			yzoomrate = (float) window_height * yimagesize / height;
 		} else {
-			xzoomrate = ximagesize;
-			yzoomrate = yimagesize;
+			xzoomrate = (float) ximagesize;
+			yzoomrate = (float) yimagesize;
 		}
 		image = myimage;
 		myimage = zoom(image, xzoomrate, yzoomrate, verbose);
 		freeImage(image);
 	} else
-		xzoomrate = yzoomrate = 100;
+		xzoomrate = yzoomrate = 100.0;
 #else
-	xzoomrate = ximagesize;
-	yzoomrate = yimagesize;
+	xzoomrate = (float) ximagesize;
+	yzoomrate = (float) yimagesize;
 	state.width = window_width;
 	state.height = window_height;
 	image_zoomratio(&state, &xzoomrate, &yzoomrate, zoomflag,
@@ -1464,7 +1509,7 @@ static void
 print_usage(name)
 	char	*name;
 {
-	fprintf(stderr, "Usage: %s [-cirV] [-f psfile] [-p paper] "
+	fprintf(stderr, "Usage: %s [-cirVt] [-f psfile] [-p paper] "
 		"[-x xmargin] [-y ymargin] [-e encoding] mgpfile\n", name);
 	exit(0);
 }
@@ -1654,7 +1699,7 @@ print_eps(epsfp, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 	char line[BUFSIZ];
 	int x1, y1, x2, y2, height, width, sheight, swidth;
 	int xpos = 0;
-	int xzoomrate, yzoomrate;
+	float xzoomrate, yzoomrate;
 	double xscale, yscale;
 	int noboundingbox;
 	struct render_state state;
@@ -1711,8 +1756,8 @@ print_eps(epsfp, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 		break;
 	}
 #else
-	xzoomrate = ximagesize;
-	yzoomrate = yimagesize;
+	xzoomrate = (float) ximagesize;
+	yzoomrate = (float) yimagesize;
 	state.width = window_width;
 	state.height = window_height;
 	image_zoomratio(&state, &xzoomrate, &yzoomrate, zoomflag,
@@ -1774,7 +1819,7 @@ main(argc, argv)
 	/* set default paper size */
 	paper = findpaper(DEFAULT_PAPER_SIZE);
 
-	while ((opt = getopt(argc, argv, "ciVrf:x:y:p:e:")) != -1) {
+	while ((opt = getopt(argc, argv, "ciVrf:x:y:p:e:Pt")) != -1) {
 		switch (opt) {
 		case 'c':
 			colorps++;
@@ -1827,6 +1872,19 @@ main(argc, argv)
 			}
 			break;
 
+		case 'P':
+			parse_debug++;
+			break;
+
+		case 't':
+#ifdef FREETYPE
+			usetfont = 1;
+#else
+			fprintf(stderr, "you can't use this function with this version\n");
+			exit(-1);
+#endif
+			break;
+
 		default:
 			print_usage(progname);
 			/*NOTREACHED*/
@@ -1878,6 +1936,12 @@ findfont(ctrl, lang, font)
 
 	if (!font[0])
 		return NULL;
+
+#ifdef FREETYPE
+	if (lang == ASCII && usetfont)
+		if ((p = findttfont(ctrl, lang, font))) goto found;
+#endif
+
 	for (p = fontmap; 0 <= p->ctrl; p++) {
 		if (p->ctrl != ctrl || p->lang != lang)
 			continue;
@@ -1938,11 +2002,12 @@ fontstring(font)
 }
 
 static void
-loadfont(font)
+loadfont(font, name)
 	struct fontmap *font;
+	char *name;
 {
 	if (!font) {
-		fprintf(fp, "unable to find proper font\n");
+		fprintf(stderr, "unable to find proper font for %s\n", name);
 		exit(1);
 	}
 
@@ -1978,3 +2043,86 @@ cleanup(int sig)
 {
 	/* dummy */
 }
+
+#ifdef FREETYPE
+/* 
+	get fontname from ttffile for ttfps 
+ */
+static struct fontmap *
+findttfont(ctrl, lang, font)
+	int ctrl;
+	int lang;
+	char *font;
+{
+	TT_Error	error;
+	TT_Face		face; 
+	TT_Face_Properties	properties;
+	u_short		platform, encoding, language, id;
+	u_short		string_len;
+	struct 		fontmap	*pfont;
+	char		*string;
+	char		pathname[256];
+	int			i, n, found = 0;
+	static TT_Engine	engine;
+	static int	initdone = 0, findex = 0; 
+	static struct fontmap	**loadedfont;
+
+	if (!font) return NULL;
+
+	/* already registed? */
+	for (i = 0; i < findex; i ++){
+		if (!strcmp(font, loadedfont[i]->font)) 
+			return loadedfont[i];
+	}
+
+	if (!initdone){
+		/* Initialize engine */
+		if ((error = TT_Init_FreeType(&engine))){
+			fprintf(stderr, "Error while initializing engine, code = %d.\n",
+				(int)error);
+				exit(-1);
+		}
+		initdone ++;
+	}
+	error = TT_Open_Face(engine, font, &face);
+	if (error) {
+		/* add freetypefontdir and try again */
+		snprintf(pathname, sizeof(pathname), "%s/%s", freetypefontdir, font);
+		error = TT_Open_Face(engine, pathname, &face);
+		if (error) return NULL;
+	}
+
+	TT_Get_Face_Properties(face, &properties);
+	n = properties.num_Names;
+	for (i = 0; i < n; i++) {
+		TT_Get_Name_ID(face, i, &platform, &encoding, &language, &id);
+		TT_Get_Name_String(face, i, &string, &string_len);
+		/* this check may be weak.. */
+		if (*string && id == 6){
+			found = 1;
+			break;
+		}
+	}
+	if (!found) return NULL;
+
+	/* create fontmap */
+	pfont = (struct fontmap *)malloc(sizeof(struct fontmap));
+	pfont->ctrl = ctrl;
+	pfont->lang = lang;
+	pfont->font = font;
+	pfont->psfont = (char *)malloc(string_len +1);
+	strncpy(pfont->psfont, string, string_len);
+	pfont->fontid = 0;
+	pfont->loaded = 0;
+
+	/* regist loaded font */
+	if (!findex)
+		loadedfont = (struct fontmap **)malloc(sizeof(struct fontmap *));
+	else
+		loadedfont = (struct fontmap **)realloc(loadedfont, 
+								sizeof(struct fontmap *) * (findex +1));
+	loadedfont[findex++] = pfont;
+
+	return pfont;
+}
+#endif /* FREETYPE */
