@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: tfont.c,v 1.30 1999/05/15 15:58:08 nishida Exp $
+ * $Id: tfont.c,v 1.34 2000/03/02 18:24:52 nishida Exp $
  */
 
 #include "mgp.h"
@@ -123,7 +123,6 @@ tfc_free(tfc)
 {
 	TFC_DELHASH(tfc);
 	TFC_DELLRU(tfc);
-	XDestroyImage(tfc->image);
 	free(tfc->fontname);
 	free(tfc->dbitmap);
 	free(tfc);
@@ -285,14 +284,21 @@ tfc_setfont(name, charset16)
 			if (!freetypefont0 || !freetypefont0[0])
 				continue;
 			fontname = freetypefont0;
+			/* already loaded? */
+			for (i = 0; i < tffontcnt; i ++) {
+				if (!strcmp(fontname, tfloadedfont[i])){
+					tfcuridx = i;
+					if (verbose) 
+						fprintf(stderr, "font \"%s\" already loaded: %d\n", 
+							fontname, tffontcnt);
+					goto cached;
+				}
+			}
 			break;
 		}
 
 		/* Load face */
-		if (verbose) {
-			fprintf(stderr, "trying to open font \"%s\"\n",
-				fontname);
-		}
+		if (verbose) fprintf(stderr, "trying to open font \"%s\"\n", fontname);
 		error = TT_Open_Face(engine, fontname, &tface);
 		if (!error)
 			break;
@@ -351,6 +357,7 @@ tfc_setfont(name, charset16)
 		goto fail2;
 	}
 
+cached:
 	if (charset16)
 		tf_mcurname = tfloadedfont[tfcuridx];
 	else
@@ -490,11 +497,10 @@ tfc_alloc(code, size, fontname, registry)
 		TTCEIL(metrics.bbox.yMax - metrics.bbox.yMin - metrics.bearingY) >> 6;
 	tfc->ascent = bitmap.rows - tfc->descent;
 	tfc->xoff = metrics.bearingX >> 6;
-	tfc->fontname = strdup(tf_curname);
+	tfc->fontname = strdup(fontname);
 	tfc->charlen = metrics.advance >> 6;
 	tfc->ref = 0;
 	tfc->dbitmap = bitmap.bitmap;
-	tfc->image = xim;
 
 	TFC_INSHASH(tfc);
 	TFC_INSLRU(tfc);
